@@ -1,5 +1,7 @@
+#include <stdio.h>
+#include <stdlib.h>
+#include <gmp.h>
 #include "field.h"
-#include "utils.h"
 
 /* returns recommended window size.  n is exponent. */
 static int optimal_pow_window_size(mpz_ptr n)
@@ -250,31 +252,11 @@ void field_set_nqr(field_ptr f, element_t nqr)
 
 void field_gen_nqr(field_ptr f)
 {
-    //use a deterministic random number generator
-    //so that the quadratic nonresidue is the same every time
-    /* Changed my mind:
-     * it is up to the application to make sure it records the
-     * quadratic nonresidue if necessary
-     */
-    /*
-    gmp_randstate_t rs;
-    void internal_random(mpz_t z, mpz_t limit, void *data)
-    {
-	UNUSED_VAR(data);
-	mpz_urandomm(z, rs, limit);
-    }
-    gmp_randinit_default(rs);
-    random_push(internal_random, NULL);
-    */
-
     f->nqr = malloc(sizeof(element_t));
     element_init(f->nqr, f);
     do {
 	element_random(f->nqr);
     } while (element_is_sqr(f->nqr));
-    /*
-    random_pop();
-    */
 }
 
 element_ptr field_get_nqr(field_ptr f)
@@ -313,3 +295,13 @@ void field_init(field_ptr f)
     f->pow_mpz = generic_pow_mpz;
     f->mul_si = generic_mul_si;
 }
+
+void field_clear(field_ptr f)
+{
+    if (f->nqr) {
+	element_clear(f->nqr);
+	free(f->nqr);
+    }
+    mpz_clear(f->order);
+    f->field_clear(f);
+}
