@@ -1,16 +1,23 @@
 #include <assert.h>
 #include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
 #include <gmp.h>
 #include "fops.h"
 #include "symtab.h"
+#include "darray.h"
+#include "field.h"
 #include "poly.h"
 #include "fieldquadratic.h"
 #include "hilbert.h"
+#include "darray.h"
 #include "mnt.h"
+#include "curve.h"
 #include "pairing.h"
 #include "d_param.h"
 #include "param.h"
 #include "tracker.h"
+#include "utils.h"
 
 struct mnt_pairing_data_s {
     field_t Fq, Fqx, Fqk;
@@ -29,7 +36,6 @@ struct even_mnt_pairing_data_s {
     element_t nqrinv, nqrinv2;
     mpz_t tateexp;
     int k;
-    fieldmap mapbase;
     element_ptr *xpowq;
 };
 typedef struct even_mnt_pairing_data_s even_mnt_pairing_data_t[1];
@@ -125,6 +131,39 @@ void d_param_inp_generic (d_param_ptr p, fetch_ops_t fops, void *ctx)
     symtab_clear(tab);
 }
 
+static inline void d_miller_evalfn_odd_k(element_t e0, element_t e1,
+	element_t a, element_t b, element_t c,
+	element_t Qx, element_t Qy, fieldmap mapbase)
+{
+    //TODO: use poly_mul_constant?
+    mapbase(e0, a);
+    element_mul(e0, e0, Qx);
+    mapbase(e1, b);
+    element_mul(e1, e1, Qy);
+    element_add(e0, e0, e1);
+    mapbase(e1, c);
+    element_add(e0, e0, e1);
+}
+
+static inline void d_miller_evalfn(element_t e0,
+	element_t a, element_t b, element_t c,
+	element_t Qx, element_t Qy)
+{
+    //a, b, c are in Fq
+    //point Q is (Qx, Qy * sqrt(nqr)) where nqr is used to construct
+    //the quadratic field extension Fqk of Fqd
+    element_ptr re_out = fi_re(e0);
+    element_ptr im_out = fi_im(e0);
+
+    int i;
+    int d = 3; //TODO: this only valid for k = 6
+    for (i=0; i<d; i++) {
+	element_mul(polymod_coeff(re_out, i), polymod_coeff(Qx, i), a);
+	element_mul(polymod_coeff(im_out, i), polymod_coeff(Qy, i), b);
+    }
+    element_add(polymod_coeff(re_out, 0), polymod_coeff(re_out, 0), c);
+}
+
 //assumes P is in the base field, Q in some field extension
 static void cc_miller(element_t res, mpz_t q, point_t P,
 	element_ptr Qx, element_ptr Qy, fieldmap mapbase)
@@ -185,14 +224,7 @@ static void cc_miller(element_t res, mpz_t q, point_t P,
 	element_add(c, c, t0);
 	element_neg(c, c);
 
-	//TODO: use poly_mul_constant?
-	mapbase(e0, a);
-	element_mul(e0, e0, Qx);
-	mapbase(e1, b);
-	element_mul(e1, e1, Qy);
-	element_add(e0, e0, e1);
-	mapbase(e1, c);
-	element_add(e0, e0, e1);
+	d_miller_evalfn_odd_k(e0, e1, a, b, c, Qx, Qy, mapbase);
 	element_mul(e, e, e0);
     }
 
@@ -230,13 +262,7 @@ static void cc_miller(element_t res, mpz_t q, point_t P,
 	element_add(c, c, t0);
 	element_neg(c, c);
 
-	mapbase(e0, a);
-	element_mul(e0, e0, Qx);
-	mapbase(e1, b);
-	element_mul(e1, e1, Qy);
-	element_add(e0, e0, e1);
-	mapbase(e1, c);
-	element_add(e0, e0, e1);
+	d_miller_evalfn_odd_k(e0, e1, a, b, c, Qx, Qy, mapbase);
 	element_mul(e, e, e0);
     }
 
@@ -282,7 +308,6 @@ static void cc_miller(element_t res, mpz_t q, point_t P,
     element_clear(c);
     element_clear(t0);
     element_clear(e0);
-    element_clear(e1);
 }
 
 static void cc_tatepower(element_ptr out, element_ptr in, pairing_t pairing)
@@ -355,7 +380,7 @@ static void trace(element_ptr out, element_ptr in, pairing_ptr pairing)
     point_clear(q);
 }
 
-static void pairing_init_c_param_odd_k(pairing_t pairing, d_param_t param)
+static void pairing_init_d_param_odd_k(pairing_t pairing, d_param_t param)
 {
     mnt_pairing_data_ptr p;
     element_t a, b;
@@ -402,14 +427,14 @@ static void pairing_init_c_param_odd_k(pairing_t pairing, d_param_t param)
     mpz_clear(z);
     p->k = param->k;
     pairing->GT = p->Fqk;
-    //pairing->phi = trace;
+    pairing->phi = trace;
 
     element_clear(a);
     element_clear(b);
 }
 
 static void cc_miller_no_denom_proj(element_t res, mpz_t q, point_t P,
-	element_ptr Qx, element_ptr Qy, fieldmap mapbase)
+	element_ptr Qx, element_ptr Qy)
 {
     int m;
     element_t v;
@@ -418,7 +443,7 @@ static void cc_miller_no_denom_proj(element_t res, mpz_t q, point_t P,
     common_curve_ptr cc = P->curve->data;
     element_t t0, t1;
     element_ptr t2 = a, t3 = b, t4 = c;
-    element_t e0, e1;
+    element_t e0;
     element_t z, z2;
 
     void proj_double(void)
@@ -427,29 +452,38 @@ static void cc_miller_no_denom_proj(element_t res, mpz_t q, point_t P,
 	element_ptr y = Z->y;
 	//t0 = 3x^2 + (cc->a) z^4
 	element_square(t0, x);
-	element_mul_si(t0, t0, 3);
+	//element_mul_si(t0, t0, 3);
+	element_double(t1, t0);
+	element_add(t0, t0, t1);
 	element_square(t1, z2);
 	element_mul(t1, t1, cc->a);
 	element_add(t0, t0, t1);
 
 	//z_out = 2 y z
 	element_mul(z, y, z);
-	element_mul_si(z, z, 2);
+	//element_mul_si(z, z, 2);
+	element_double(z, z);
 	element_square(z2, z);
 
 	//t1 = 4 x y^2
 	element_square(t2, y);
 	element_mul(t1, x, t2);
-	element_mul_si(t1, t1, 4);
+	//element_mul_si(t1, t1, 4);
+	element_double(t1, t1);
+	element_double(t1, t1);
 
 	//x_out = t0^2 - 2 t1
-	element_mul_si(t3, t1, 2);
+	//element_mul_si(t3, t1, 2);
+	element_double(t3, t1);
 	element_square(x, t0);
 	element_sub(x, x, t3);
 
 	//t2 = 8y^4
 	element_square(t2, t2);
-	element_mul_si(t2, t2, 8);
+	//element_mul_si(t2, t2, 8);
+	element_double(t2, t2);
+	element_double(t2, t2);
+	element_double(t2, t2);
 
 	//y_out = t0(t1 - x_out) - t2
 	element_sub(t1, t1, x);
@@ -514,16 +548,6 @@ static void cc_miller_no_denom_proj(element_t res, mpz_t q, point_t P,
 	}
     }
 
-    void do_vertical(void)
-    {
-	if (point_is_inf(Z)) {
-	    return;
-	}
-	mapbase(e0, Z->x);
-	element_sub(e0, Qx, e0);
-	element_mul(v, v, e0);
-    }
-
     void do_tangent(void)
     {
 	element_ptr Zx = Z->x;
@@ -536,7 +560,9 @@ static void cc_miller_no_denom_proj(element_t res, mpz_t q, point_t P,
 	element_square(a, z2);
 	element_mul(a, a, cc->a);
 	element_square(b, Zx);
-	element_mul_si(b, b, 3);
+	//element_mul_si(b, b, 3);
+	element_double(t0, b);
+	element_add(b, b, t0);
 	element_add(a, a, b);
 	element_neg(a, a);
 
@@ -551,14 +577,7 @@ static void cc_miller_no_denom_proj(element_t res, mpz_t q, point_t P,
 	element_add(c, c, t0);
 	element_neg(c, c);
 
-	//TODO: use poly_mul_constant?
-	mapbase(e0, a);
-	element_mul(e0, e0, Qx);
-	mapbase(e1, b);
-	element_mul(e1, e1, Qy);
-	element_add(e0, e0, e1);
-	mapbase(e1, c);
-	element_add(e0, e0, e1);
+	d_miller_evalfn(e0, a, b, c, Qx, Qy);
 	element_mul(v, v, e0);
     }
 
@@ -586,13 +605,7 @@ static void cc_miller_no_denom_proj(element_t res, mpz_t q, point_t P,
 	element_mul(c, Ay, Bx);
 	element_sub(c, t0, c);
 
-	mapbase(e0, a);
-	element_mul(e0, e0, Qx);
-	mapbase(e1, b);
-	element_mul(e1, e1, Qy);
-	element_add(e0, e0, e1);
-	mapbase(e1, c);
-	element_add(e0, e0, e1);
+	d_miller_evalfn(e0, a, b, c, Qx, Qy);
 	element_mul(v, v, e0);
     }
 
@@ -602,7 +615,6 @@ static void cc_miller_no_denom_proj(element_t res, mpz_t q, point_t P,
     element_init(t0, a->field);
     element_init(t1, a->field);
     element_init(e0, res->field);
-    element_init(e1, res->field);
     element_init(z, a->field);
     element_init(z2, a->field);
     element_set1(z);
@@ -638,13 +650,12 @@ static void cc_miller_no_denom_proj(element_t res, mpz_t q, point_t P,
     element_clear(t0);
     element_clear(t1);
     element_clear(e0);
-    element_clear(e1);
     element_clear(z);
     element_clear(z2);
 }
 
-static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
-	element_ptr Qx, element_ptr Qy, fieldmap mapbase)
+static void cc_miller_no_denom_affine(element_t res, mpz_t q, point_t P,
+	element_ptr Qx, element_ptr Qy)
 {
     int m;
     element_t v;
@@ -653,14 +664,16 @@ static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
     const common_curve_ptr cc = P->curve->data;
     const element_ptr cca = cc->a;
     element_t t0;
-    element_t e0, e1;
+    element_t e0;
 
+    /* TODO: when exactly is this not needed?
     void do_vertical(void)
     {
 	mapbase(e0, Z->x);
 	element_sub(e0, Qx, e0);
 	element_mul(v, v, e0);
     }
+    */
 
     void do_tangent(void)
     {
@@ -682,14 +695,7 @@ static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
 	element_add(c, c, t0);
 	element_neg(c, c);
 
-	//TODO: use poly_mul_constant?
-	mapbase(e0, a);
-	element_mul(e0, e0, Qx);
-	mapbase(e1, b);
-	element_mul(e1, e1, Qy);
-	element_add(e0, e0, e1);
-	mapbase(e1, c);
-	element_add(e0, e0, e1);
+	d_miller_evalfn(e0, a, b, c, Qx, Qy);
 	element_mul(v, v, e0);
     }
 
@@ -712,13 +718,7 @@ static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
 	element_add(c, c, t0);
 	element_neg(c, c);
 
-	mapbase(e0, a);
-	element_mul(e0, e0, Qx);
-	mapbase(e1, b);
-	element_mul(e1, e1, Qy);
-	element_add(e0, e0, e1);
-	mapbase(e1, c);
-	element_add(e0, e0, e1);
+	d_miller_evalfn(e0, a, b, c, Qx, Qy);
 	element_mul(v, v, e0);
     }
 
@@ -727,7 +727,6 @@ static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
     element_init(c, P->curve->field);
     element_init(t0, P->curve->field);
     element_init(e0, res->field);
-    element_init(e1, res->field);
 
     element_init(v, res->field);
     point_init(Z, P->curve);
@@ -760,7 +759,6 @@ static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
     element_clear(c);
     element_clear(t0);
     element_clear(e0);
-    element_clear(e1);
 }
 
 static void cc_tatepower_even_k(element_ptr out, element_ptr in, pairing_t pairing)
@@ -814,24 +812,27 @@ static void cc_tatepower_even_k(element_ptr out, element_ptr in, pairing_t pairi
     }
 }
 
+static void (*cc_miller_no_denom_fn)(element_t res, mpz_t q, point_t P,
+	element_ptr Qx, element_ptr Qy);
+
 static void cc_pairing_even_k(element_ptr out, element_ptr in1, element_ptr in2,
 	pairing_t pairing)
 {
     point_ptr Qbase = in2->data;
-    element_t x, y;
+    element_t Qx, Qy;
     even_mnt_pairing_data_ptr p = pairing->data;
 
-    element_init(x, out->field);
-    element_init(y, out->field);
+    element_init(Qx, p->Fqd);
+    element_init(Qy, p->Fqd);
     //map from twist: (x, y) --> (v^-1 x, v^-(3/2) y)
     //where v is the quadratic nonresidue used to construct the twist
-    element_mul(fi_re(x), Qbase->x, p->nqrinv);
+    element_mul(Qx, Qbase->x, p->nqrinv);
     //v^-3/2 = v^-2 * v^1/2
-    element_mul(fi_im(y), Qbase->y, p->nqrinv2);
-    cc_miller_no_denom(out, pairing->r, in1->data, x, y, p->mapbase);
+    element_mul(Qy, Qbase->y, p->nqrinv2);
+    cc_miller_no_denom_fn(out, pairing->r, in1->data, Qx, Qy);
     cc_tatepower_even_k(out, out, pairing);
-    element_clear(x);
-    element_clear(y);
+    element_clear(Qx);
+    element_clear(Qy);
 }
 
 static int cc_is_almost_coddh_even_k(element_ptr a, element_ptr b,
@@ -844,10 +845,10 @@ static int cc_is_almost_coddh_even_k(element_ptr a, element_ptr b,
     element_t dx, dy;
     even_mnt_pairing_data_ptr p = pairing->data;
 
-    element_init(cx, p->Fqk);
-    element_init(cy, p->Fqk);
-    element_init(dx, p->Fqk);
-    element_init(dy, p->Fqk);
+    element_init(cx, p->Fqd);
+    element_init(cy, p->Fqd);
+    element_init(dx, p->Fqd);
+    element_init(dy, p->Fqd);
 
     element_init(t0, pairing->GT);
     element_init(t1, pairing->GT);
@@ -856,14 +857,14 @@ static int cc_is_almost_coddh_even_k(element_ptr a, element_ptr b,
     point_ptr Pd = d->data;
     //map from twist: (x, y) --> (v^-1 x, v^-(3/2) y)
     //where v is the quadratic nonresidue used to construct the twist
-    element_mul(fi_re(cx), Pc->x, p->nqrinv);
-    element_mul(fi_re(dx), Pd->x, p->nqrinv);
+    element_mul(cx, Pc->x, p->nqrinv);
+    element_mul(dx, Pd->x, p->nqrinv);
     //v^-3/2 = v^-2 * v^1/2
-    element_mul(fi_im(cy), Pc->y, p->nqrinv2);
-    element_mul(fi_im(dy), Pd->y, p->nqrinv2);
+    element_mul(cy, Pc->y, p->nqrinv2);
+    element_mul(dy, Pd->y, p->nqrinv2);
 
-    cc_miller_no_denom(t0, pairing->r, a->data, dx, dy, p->mapbase);
-    cc_miller_no_denom(t1, pairing->r, b->data, cx, cy, p->mapbase);
+    cc_miller_no_denom_fn(t0, pairing->r, a->data, dx, dy);
+    cc_miller_no_denom_fn(t1, pairing->r, b->data, cx, cy);
     cc_tatepower_even_k(t0, t0, pairing);
     cc_tatepower_even_k(t1, t1, pairing);
     element_mul(t2, t0, t1);
@@ -878,6 +879,10 @@ static int cc_is_almost_coddh_even_k(element_ptr a, element_ptr b,
 	    res = 1;
 	}
     }
+    element_clear(cx);
+    element_clear(cy);
+    element_clear(dx);
+    element_clear(dy);
     element_clear(t0);
     element_clear(t1);
     element_clear(t2);
@@ -937,7 +942,201 @@ static void even_k_trace(element_t out, element_t in, pairing_ptr pairing)
     point_clear(r);
 }
 
-static void pairing_init_c_param_even_k(pairing_t pairing, d_param_t param)
+static void d_pairing_option_set(pairing_t pairing, char *key, char *value)
+{
+    UNUSED_VAR(pairing);
+    if (!strcmp(key, "coord")) {
+	if (!strcmp(value, "projective")) {
+	    cc_miller_no_denom_fn = cc_miller_no_denom_proj;
+	} else if (!strcmp(value, "affine")) {
+	    cc_miller_no_denom_fn = cc_miller_no_denom_affine;
+	}
+    }
+}
+
+struct pp_coeff_s {
+    element_t a;
+    element_t b;
+    element_t c;
+};
+typedef struct pp_coeff_s pp_coeff_t[1];
+typedef struct pp_coeff_s *pp_coeff_ptr;
+
+static void d_pairing_pp_init_even_k(pairing_pp_t p, element_ptr in1, pairing_t pairing)
+{
+    point_ptr P = in1->data;
+    const common_curve_ptr cc = P->curve->data;
+    point_t Z;
+    int m;
+    even_mnt_pairing_data_ptr info = pairing->data;
+    element_t t0;
+    element_t a, b, c;
+    field_ptr Fq = info->Fq;
+    pp_coeff_t *coeff;
+    mpz_ptr q = pairing->r;
+    pp_coeff_ptr pp;
+
+    void store_abc(void)
+    {
+	element_init(pp->a, Fq);
+	element_init(pp->b, Fq);
+	element_init(pp->c, Fq);
+	element_set(pp->a, a);
+	element_set(pp->b, b);
+	element_set(pp->c, c);
+	pp++;
+    }
+
+    void do_tangent(void)
+    {
+	//a = -slope_tangent(Z.x, Z.y);
+	//b = 1;
+	//c = -(Z.y + a * Z.x);
+	//but we multiply by 2*Z.y to avoid division
+
+	//a = -Zx * (3 Zx + twicea_2) - a_4;
+	//Common curves: a2 = 0 (and cc->a is a_4), so
+	//a = -(3 Zx^2 + cc->a)
+	//b = 2 * Zy
+	//c = -(2 Zy^2 + a Zx);
+	const element_ptr Zx = Z->x;
+	const element_ptr Zy = Z->y;
+
+	element_square(a, Zx);
+	element_double(t0, a);
+	element_add(a, a, t0);
+	element_add(a, a, cc->a);
+	element_neg(a, a);
+
+	element_add(b, Zy, Zy);
+
+	element_mul(t0, b, Zy);
+	element_mul(c, a, Zx);
+	element_add(c, c, t0);
+	element_neg(c, c);
+
+	store_abc();
+    }
+
+    void do_line(void)
+    {
+	//a = -(B.y - A.y) / (B.x - A.x);
+	//b = 1;
+	//c = -(A.y + a * A.x);
+	//but we'll multiply by B.x - A.x to avoid division
+
+	const element_ptr Ax = Z->x;
+	const element_ptr Ay = Z->y;
+	const element_ptr Bx = P->x;
+	const element_ptr By = P->y;
+
+	element_sub(b, Bx, Ax);
+	element_sub(a, Ay, By);
+	element_mul(t0, b, Ay);
+	element_mul(c, a, Ax);
+	element_add(c, c, t0);
+	element_neg(c, c);
+
+	store_abc();
+    }
+
+    point_init(Z, P->curve);
+    point_set(Z, P);
+
+    element_init(t0, Fq);
+    element_init(a, Fq);
+    element_init(b, Fq);
+    element_init(c, Fq);
+
+    m = mpz_sizeinbase(q, 2) - 2;
+    p->data = malloc(sizeof(pp_coeff_t) * 2 * m);
+    coeff = (pp_coeff_t *) p->data;
+    pp = coeff[0];
+
+    for(;;) {
+	do_tangent();
+
+	if (!m) break;
+
+	point_double(Z, Z);
+	if (mpz_tstbit(q, m)) {
+	    do_line();
+	    point_add(Z, Z, P);
+	}
+	m--;
+    }
+
+    element_clear(t0);
+    element_clear(a);
+    element_clear(b);
+    element_clear(c);
+    point_clear(Z);
+}
+
+static void d_pairing_pp_clear_even_k(pairing_pp_t p)
+{
+    //TODO: better to store a sentinel value in p->data?
+    mpz_ptr q = p->pairing->r;
+    int m = mpz_sizeinbase(q, 2) + mpz_popcount(q) - 3;
+    int i;
+    pp_coeff_t *coeff = (pp_coeff_t *) p->data;
+    pp_coeff_ptr pp;
+    for (i=0; i<m; i++) {
+	pp = coeff[i];
+	element_clear(pp->a);
+	element_clear(pp->b);
+	element_clear(pp->c);
+    }
+    free(p->data);
+}
+
+static void d_pairing_pp_apply_even_k(element_ptr out, element_ptr in2, pairing_pp_t p)
+{
+    mpz_ptr q = p->pairing->r;
+    even_mnt_pairing_data_ptr info = p->pairing->data;
+    int m = mpz_sizeinbase(q, 2) - 2;
+    pp_coeff_t *coeff = (pp_coeff_t *) p->data;
+    pp_coeff_ptr pp = coeff[0];
+    point_ptr Qbase = in2->data;
+    element_t e0;
+    element_t Qx, Qy;
+    element_t v;
+    element_init_GT(e0, p->pairing);
+    element_init_GT(v, p->pairing);
+    element_init(Qx, info->Fqd);
+    element_init(Qy, info->Fqd);
+
+    //map from twist: (x, y) --> (v^-1 x, v^-(3/2) y)
+    //where v is the quadratic nonresidue used to construct the twist
+    element_mul(Qx, Qbase->x, info->nqrinv);
+    //v^-3/2 = v^-2 * v^1/2
+    element_mul(Qy, Qbase->y, info->nqrinv2);
+
+    element_set1(out);
+    for(;;) {
+	d_miller_evalfn(e0, pp->a, pp->b, pp->c, Qx, Qy);
+	element_mul(out, out, e0);
+	pp++;
+
+	if (!m) break;
+
+	if (mpz_tstbit(q, m)) {
+	    d_miller_evalfn(e0, pp->a, pp->b, pp->c, Qx, Qy);
+	    element_mul(out, out, e0);
+	    pp++;
+	}
+	m--;
+	element_square(out, out);
+    }
+    cc_tatepower_even_k(out, out, p->pairing);
+
+    element_clear(e0);
+    element_clear(Qx);
+    element_clear(Qy);
+    element_clear(v);
+}
+
+static void pairing_init_d_param_even_k(pairing_t pairing, d_param_t param)
 {
     even_mnt_pairing_data_ptr p;
     element_t a, b;
@@ -1003,9 +1202,7 @@ static void pairing_init_c_param_even_k(pairing_t pairing, d_param_t param)
 	mpz_divexact(p->tateexp, p->tateexp, pairing->r);
     }
 
-    p->mapbase = Fq_to_Fqd_to_Fqk;
-
-    cc_init_map_curve(p->Eqk, p->Eq, p->Fqk, p->mapbase);
+    cc_init_map_curve(p->Eqk, p->Eq, p->Fqk, Fq_to_Fqd_to_Fqk);
     cc_init_map_curve(p->Etwist, p->Eq, p->Fqd, element_field_to_polymod);
     twist_curve(p->Etwist);
     element_init(p->nqrinv, p->Fqd);
@@ -1025,17 +1222,23 @@ static void pairing_init_c_param_even_k(pairing_t pairing, d_param_t param)
     pairing->GT = p->Fqk;
     pairing->phi = even_k_trace;
 
+    cc_miller_no_denom_fn = cc_miller_no_denom_affine;
+    pairing->option_set = d_pairing_option_set;
+    pairing->pp_init = d_pairing_pp_init_even_k;
+    pairing->pp_clear = d_pairing_pp_clear_even_k;
+    pairing->pp_apply = d_pairing_pp_apply_even_k;
+
     element_clear(a);
     element_clear(b);
 }
 
-void pairing_init_c_param(pairing_t pairing, d_param_t param)
+void pairing_init_d_param(pairing_t pairing, d_param_t param)
 {
     if (0) {
-	pairing_init_c_param_odd_k(pairing, param);
+	pairing_init_d_param_odd_k(pairing, param);
     } else {
-	if (param->k % 2) pairing_init_c_param_odd_k(pairing, param);
-	else pairing_init_c_param_even_k(pairing, param);
+	if (param->k % 2) pairing_init_d_param_odd_k(pairing, param);
+	else pairing_init_d_param_even_k(pairing, param);
     }
 }
 
